import struct

import six

from .flags import (
    TAC_PLUS_AUTHEN_SVC_LOGIN, TAC_PLUS_ACCT_STATUS_SUCCESS,
    TAC_PLUS_ACCT_STATUS_ERROR, TAC_PLUS_ACCT_STATUS_FOLLOW,
    TAC_PLUS_VIRTUAL_PORT, TAC_PLUS_VIRTUAL_REM_ADDR
)


class TACACSAccountingStart(object):

    def __init__(self, username, flags, authen_method, priv_lvl, authen_type,
                 arguments, rem_addr=TAC_PLUS_VIRTUAL_REM_ADDR,
                 port=TAC_PLUS_VIRTUAL_PORT):
        self.username = username
        self.flags = flags
        self.authen_method = authen_method
        self.priv_lvl = priv_lvl
        self.authen_type = authen_type
        self.service = TAC_PLUS_AUTHEN_SVC_LOGIN
        self.arguments = arguments
        self.rem_addr = rem_addr
        self.port = port

    @property
    def packed(self):
        #  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8
        # +----------------+----------------+----------------+----------------+
        # |      flags     |  authen_method |    priv_lvl    |  authen_type   |
        # +----------------+----------------+----------------+----------------+
        # | authen_service |    user_len    |    port_len    |  rem_addr_len  |
        # +----------------+----------------+----------------+----------------+
        # |    arg_cnt     |   arg_1_len    |   arg_2_len    |      ...       |
        # +----------------+----------------+----------------+----------------+
        # |   arg_N_len    |    user ...
        # +----------------+----------------+----------------+----------------+
        # |   port ...
        # +----------------+----------------+----------------+----------------+
        # |   rem_addr ...
        # +----------------+----------------+----------------+----------------+
        # |   arg_1 ...
        # +----------------+----------------+----------------+----------------+
        # |   arg_2 ...
        # +----------------+----------------+----------------+----------------+
        # |   ...
        # +----------------+----------------+----------------+----------------+
        # |   arg_N ...
        # +----------------+----------------+----------------+----------------+
        username = six.b(self.username)
        rem_addr = six.b(self.rem_addr)
        port = six.b(self.port)
        arguments = self.arguments
        body = struct.pack(
            'B' * 9,
            self.flags,
            self.authen_method,
            self.priv_lvl,
            self.authen_type,
            self.service,
            len(username),
            len(port),
            len(rem_addr),
            len(arguments),
        )
        for value in arguments:
            body += struct.pack('B', len(value))
        for value in (username, port, rem_addr):
            body += struct.pack('%ds' % len(value), value)
        for value in arguments:
            body += struct.pack('%ds' % len(value), value)
        return body

    def __str__(self):
        args = ', '.join([x.decode('utf-8') for x in self.arguments])
        return ', '.join([
            'args: %s' % args,
            'args_cnt: %d' % len(self.arguments),
            'authen_method: %s' % self.authen_method,
            'authen_type: %s' % self.authen_type,
            'authen_service: %s' % self.service,
            'flags: %s' % self.flags,
            'port_len: %d' % len(self.port),
            'port: %s' % self.port,
            'priv_lvl: %s' % self.priv_lvl,
            'rem_addr_len: %d' % len(self.rem_addr),
            'user: %s' % self.username,
            'user_len: %d' % len(self.username),
        ])


class TACACSAccountingReply(object):
    def __init__(self, status, server_msg, data):
        self.status = status
        self.server_msg = server_msg
        self.data = data
        self.flags = None
        self.arguments = []

    @classmethod
    def unpacked(cls, raw):
        #  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8
        # +----------------+----------------+----------------+----------------+
        # |         server_msg len          |            data_len             |
        # +----------------+----------------+----------------+----------------+
        # |     status     |         server_msg ...
        # +----------------+----------------+----------------+----------------+
        # |     data ...
        # +----------------+

        # B = unsigned char
        # !H = network-order (big-endian) unsigned short
        raw = six.BytesIO(raw)
        server_msg_len, data_len = struct.unpack('!HH', raw.read(4))
        status = struct.unpack('B', raw.read(1))[0]
        server_msg = raw.read(server_msg_len)
        data = raw.read(data_len) if data_len else b''
        return cls(status, server_msg, data)

    @property
    def valid(self):
        return self.status == TAC_PLUS_ACCT_STATUS_SUCCESS

    @property
    def error(self):
        return self.status == TAC_PLUS_ACCT_STATUS_ERROR

    @property
    def follow(self):
        return self.status == TAC_PLUS_ACCT_STATUS_FOLLOW

    @property
    def human_status(self):
        return {
            TAC_PLUS_ACCT_STATUS_SUCCESS: 'SUCCESS',
            TAC_PLUS_ACCT_STATUS_ERROR: 'ERROR',
            TAC_PLUS_ACCT_STATUS_FOLLOW: 'FOLLOW'
        }.get(self.status, 'UNKNOWN: %s' % self.status)

    def __str__(self):
        return ', '.join([
            'data: %s' % self.data,
            'data_len: %d' % len(self.data),
            'server_msg: %s' % self.server_msg,
            'server_msg_len: %d' % len(self.server_msg),
            'status: %s' % self.human_status,
        ])



import struct
from hashlib import md5

import six


def crypt(header, body_bytes, secret):
    """
    TACACS+ uses a shared secret key (known to both the client and server)
    to obfuscate the body of sent packets.  Only the packet body (not
    the header) is obfuscated.

    https://datatracker.ietf.org/doc/draft-ietf-opsawg-tacacs/?include_text=1#section-3.7

    ENCRYPTED {data} == data ^ pseudo_pad

    The pad is generated by concatenating a series of MD5 hashes (each 16
    bytes long) and truncating it to the length of the input data.

    pseudo_pad = {MD5_1 [,MD5_2 [ ... ,MD5_n]]} truncated to len(data)

    The first MD5 hash is generated by concatenating the session_id, the
    secret key, the version number and the sequence number and then
    running MD5 over that stream.  All of those input values are
    available in the packet header, except for the secret key which is a
    shared secret between the TACACS+ client and server.

    Subsequent hashes are generated by using the same input stream, but
    concatenating the previous hash value at the end of the input stream.

    MD5_1 = MD5{session_id, key, version, seq_no} MD5_2 = MD5{session_id,
    key, version, seq_no, MD5_1} ....  MD5_n = MD5{session_id, key,
    version, seq_no, MD5_n-1}

    :param header:     a TACACSHeader object
    :param body_bytes: packed bytes, i.e., `struct.pack(...)`
    :param secret:     a key used to encrypt/obfuscate packets according
                       to the TACACS+ spec

    :return:           packed bytes, i.e., `struct.pack(...)` representing the
                       obfuscated packet body
    """  # noqa

    # B = unsigned char
    # !I = network-order (big-endian) unsigned int
    body_length = len(body_bytes)
    unhashed = (
        struct.pack('!I', header.session_id) +
        six.b(secret) +
        struct.pack('B', header.version) +
        struct.pack('B', header.seq_no)
    )
    pad = hashed = md5(unhashed).digest()

    if (len(pad) < body_length):
        # remake hash, appending it to pad until pad >= header.length
        while True:
            hashed = md5(unhashed + hashed).digest()
            pad += hashed
            if len(pad) >= body_length:
                break

    pad = pad[0:(body_length)]
    pad = list(struct.unpack('B' * len(pad), pad))

    packet_body = []
    for x in struct.unpack('B' * body_length, body_bytes):
        packet_body.append(x ^ pad.pop(0))

    return struct.pack('B' * len(packet_body), *packet_body)


class TACACSPacket(object):

    def __init__(self, header, body_bytes, secret):
        """
        :param header:     a TACACSHeader object
        :param body_bytes: packed bytes, i.e., `struct.pack(...)`
        :param secret:     a key used to encrypt/obfuscate packets according
                           to the TACACS+ spec
        """
        self.header = header
        self.body_bytes = body_bytes
        self.secret = secret

    @property
    def encrypted(self):
        return self.secret is not None

    @property
    def seq_no(self):
        return self.header.seq_no

    @property
    def body(self):
        if self.encrypted:
            return self.crypt
        return self.body_bytes

    def __str__(self):
        return self.header.packed + self.body

    def __bytes__(self):
        return self.header.packed + self.body

    @property
    def crypt(self):
        return crypt(self.header, self.body_bytes, self.secret)


class TACACSHeader(object):

    def __init__(self, version, type, session_id, length, seq_no=1, flags=0):
        self.version = version
        self.type = type
        self.session_id = session_id
        self.length = length
        self.seq_no = seq_no
        self.flags = flags

    @property
    def version_max(self):
        return self.version // 0x10

    @property
    def version_min(self):
        return self.version % 0x10

    @property
    def packed(self):
        # All TACACS+ packets always begin with the following 12 byte header.
        # The header is always cleartext and describes the remainder of the
        # packet:
        # 1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8
        #
        # +----------------+----------------+----------------+----------------+
        # |major  | minor  |                |                |                |
        # |version| version|      type      |     seq_no     |   flags        |
        # +----------------+----------------+----------------+----------------+
        # |                            session_id                             |
        # +----------------+----------------+----------------+----------------+
        # |                              length                               |
        # +----------------+----------------+----------------+----------------+

        # B = unsigned char
        # !I = network-order (big-endian) unsigned int
        return struct.pack(
            'BBBB',
            self.version,
            self.type,
            self.seq_no,
            self.flags
        ) + struct.pack('!I', self.session_id) + struct.pack('!I', self.length)

    @classmethod
    def unpacked(cls, raw):
        # B = unsigned char
        # !I = network-order (big-endian) unsigned int
        raw = six.BytesIO(raw)
        raw_chars = raw.read(4)
        if raw_chars:
            version, type, seq_no, flags = struct.unpack(
                'BBBB',
                raw_chars
            )
            session_id, length = struct.unpack('!II', raw.read(8))
            return cls(version, type, session_id, length, seq_no, flags)
        else:
            raise ValueError(
                "Unable to extract data from header. Likely the TACACS+ key does not match between server and client"
            )

    def __str__(self):
        return ', '.join([
            'version: %s' % self.version,
            'type: %s' % self.type,
            'session_id: %s' % self.session_id,
            'length: %s' % self.length,
            'seq_no: %s' % self.seq_no,
            'flags: %s' % self.flags,
        ])



# Copyright (c) 2017 Ansible by Red Hat
# All Rights Reserved.

from hashlib import md5
import socket
import types
from uuid import uuid4

import pytest
import six

from tacacs_plus.flags import (
    TAC_PLUS_AUTHEN, TAC_PLUS_AUTHEN_TYPE_ASCII,
    TAC_PLUS_AUTHEN_TYPE_PAP, TAC_PLUS_AUTHEN_TYPE_CHAP,
    TAC_PLUS_PRIV_LVL_MIN, TAC_PLUS_PRIV_LVL_MAX,
    TAC_PLUS_AUTHEN_METH_TACACSPLUS,
    TAC_PLUS_ACCT_FLAG_START
)
from tacacs_plus.packet import TACACSHeader, TACACSPacket, crypt
from tacacs_plus.authentication import (
    TACACSAuthenticationStart, TACACSAuthenticationContinue, TACACSAuthenticationReply
)
from tacacs_plus.authorization import TACACSAuthorizationStart, TACACSAuthorizationReply
from tacacs_plus.accounting import TACACSAccountingStart, TACACSAccountingReply
from tacacs_plus.client import TACACSClient

AUTHENTICATE_HEADER = six.b('\xc0\x01\x01\x00\x00\x0009\x00\x00\x00')
AUTHENTICATE_HEADER_V12_1 = six.b('\xc1\x01\x01\x00\x00\x0009\x00\x00\x00')
AUTHENTICATE_HEADER_WRONG = six.b('\xf0\x01\x01\x00\x00\x0009\x00\x00\x00')
AUTHORIZE_HEADER = six.b('\xc0\x02\x01\x00\x00\x0009\x00\x00\x00')
ACCOUNT_HEADER = six.b('\xc0\x03\x01\x00\x00\x0009\x00\x00\x00')


@pytest.fixture
def fake_socket(tmpdir_factory, request):
    packets = request.node.callspec.params.get('packets')

    # write data to the "socket"; this must be an actual file, because
    # select.select()() expects a real file descriptor
    filename = str(tmpdir_factory.mktemp('fake-socket').join(str(uuid4())))
    f = open(filename, 'w')
    request.addfinalizer(f.close)

    sockobj = socket._socket.socket if six.PY3 else socket._socketobject

    class fakesocket(sockobj):
        buff = six.BytesIO()

    def _send(self, data):
        self.buff.write(data)
        return len(data)  # number of bytes sent

    sock = fakesocket()
    sock.f = six.BytesIO(packets)
    # socket.socket overrides these at instantiation time to the underlying
    # C implementation; set them here so that we can mock send() and recv()
    # calls
    sock.send = types.MethodType(_send, sock)
    sock.recv = types.MethodType(
        lambda self, _bytes: self.f.read(_bytes),
        sock
    )
    sock.fileno = types.MethodType(lambda self: f.fileno(), sock)
    return sock


@pytest.fixture
def tacacs_header():
    return TACACSHeader(
        192,  # version
        TAC_PLUS_AUTHEN,
        12345,  # session_id,
        0,  # body len
        1,  # seq_no
    )


@pytest.mark.parametrize('username', ['username', 'long-very-padded-username'])
def test_packet_body_obfuscation(tacacs_header, username):
    header = tacacs_header
    body = TACACSAuthenticationStart(
        username,
        TAC_PLUS_AUTHEN_TYPE_ASCII
    ).packed
    secret = '6f188b78-964f-4f3e-aea7-da26b7772495'
    encrypted = crypt(header, body, secret)
    assert crypt(header, encrypted, secret) == body


def test_packet_seq_no(tacacs_header):
    packet = TACACSPacket(tacacs_header, '', None)
    assert packet.seq_no == 1


def test_packet_encrypted(tacacs_header):
    assert TACACSPacket(tacacs_header, '', None).encrypted is False
    assert TACACSPacket(tacacs_header, '', 'secret').encrypted is True


@pytest.mark.parametrize('secret_key', ('secret', None))
def test_packet_raw(secret_key, tacacs_header):
    header = tacacs_header
    body = TACACSAuthenticationStart(
        'user123',
        TAC_PLUS_AUTHEN_TYPE_ASCII
    )
    header.length = len(body.packed)
    packet = TACACSPacket(header, body.packed, secret_key)
    assert isinstance(packet.header, TACACSHeader)

    if secret_key:
        assert bytes(packet) == header.packed + crypt(header, body.packed, secret_key)
    else:
        assert bytes(packet) == header.packed + body.packed


def test_header_pack(tacacs_header):
    packed = tacacs_header.packed
    unpacked = TACACSHeader.unpacked(packed)
    assert unpacked.version == 192
    assert unpacked.type == TAC_PLUS_AUTHEN
    assert unpacked.session_id == 12345
    assert unpacked.length == 0
    assert unpacked.seq_no == 1
    assert unpacked.flags == 0


# test authentication objects
def test_authentication_start_pack():
    username = 'user123'
    packed = TACACSAuthenticationStart(
        username,
        TAC_PLUS_AUTHEN_TYPE_ASCII
    ).packed
    assert packed == six.b(
        '\x01'           # tacacs_plus.flags.TAC_PLUS_AUTHEN_LOGIN \
        '\x00'           # tacacs_plus.flags.TAC_PLUS_PRIV_LVL_MIN \
        '\x01'           # tacacs_plus.flags.TAC_PLUS_AUTHEN_TYPE_ASCII \
        '\x01'           # tacacs_plus.flags.TAC_PLUS_AUTHEN_SVC_LOGIN \
        '\x07'           # username_len == 7 \
        '\x0b'           # port_len == 11 \
        '\x0d'           # rem_addr_len == 13 \
        '\x00'           # data_len == 0 \
        'user123'        # username \
        'python_tty0'    # port \
        'python_device'  # rem_addr
    )


def test_authentication_continue_pack():
    password = 'password'
    packed = TACACSAuthenticationContinue(password).packed
    assert packed == six.b(
        '\x00\x08'  # user_msg len (password length) \
        '\x00\x00'  # data_len \
        '\x00'      # flags \
        'password'
    )


@pytest.mark.parametrize('raw, state, human_status', [
    (six.b('\x01\x00\x00\x00\x00\x00'), 'valid', 'PASS'),
    (six.b('\x02\x00\x00\x00\x00\x00'), 'invalid', 'FAIL',),
    (six.b('\x07\x00\x00\x00\x00\x00'), 'error', 'ERROR'),
])
def test_authentication_reply_unpack(raw, state, human_status):
    reply = TACACSAuthenticationReply.unpacked(raw)
    assert getattr(reply, state) is True
    assert reply.human_status == human_status


def test_authentication_reply_unpack_server_msg():
    reply = TACACSAuthenticationReply.unpacked(
        six.b('\x05\x01\x00\n\x00\x00Password: ')
    )
    assert reply.getpass is True
    assert reply.server_msg == six.b('Password: ')
    assert reply.human_status == 'GETPASS'


# test authorization objects
def test_authorization_start_pack():
    username = 'user123'
    packed = TACACSAuthorizationStart(
        username, TAC_PLUS_AUTHEN_METH_TACACSPLUS,
        TAC_PLUS_PRIV_LVL_MIN, TAC_PLUS_AUTHEN_TYPE_ASCII,
        [b"service=shell", b"cmd=show", b"cmdargs=version"],
    ).packed
    assert packed == six.b(
        '\x06'              # tacacs_plus.flags.TAC_PLUS_AUTHEN_METH_TACACSPLUS \
        '\x00'              # tacacs_plus.flags.TAC_PLUS_PRIV_LVL_MIN \
        '\x01'              # tacacs_plus.flags.TAC_PLUS_AUTHEN_TYPE_ASCII \
        '\x01'              # tacacs_plus.flags.TAC_PLUS_AUTHEN_SVC_LOGIN \
        '\x07'              # username_len == 7 \
        '\x0b'              # port_len == 11 \
        '\x0d'              # rem_addr_len == 13 \
        '\x03'              # arg_len == 3
        '\x0d'              # arg_1_len == 13 (service=shell) \
        '\x08'              # arg_2_len == 8 (cmd=show) \
        '\x0f'              # arg_3_len == 14 (cmdargs=version) \
        'user123'           # username \
        'python_tty0'       # port \
        'python_device'     # rem_addr \
        'service=shell'     # arg_1 (service=shell) \
        'cmd=show'          # arg_2 (cmd=show) \
        'cmdargs=version'   # arg_3 (cmdargs=version)
    )


@pytest.mark.parametrize('raw, state, human_status', [
    (six.b('\x01\x00\x00\x00\x00\x00'), 'valid', 'PASS'),
    (six.b('\x02\x00\x00\x00\x00\x00'), 'reply', 'REPL'),
    (six.b('\x10\x00\x00\x00\x00\x00'), 'invalid', 'FAIL',),
    (six.b('\x11\x00\x00\x00\x00\x00'), 'error', 'ERROR'),
    (six.b('\x21\x00\x00\x00\x00\x00'), 'follow', 'FOLLOW'),
])
def test_authorization_reply_unpack(raw, state, human_status):
    reply = TACACSAuthorizationReply.unpacked(raw)
    assert getattr(reply, state) is True
    assert reply.human_status == human_status


def test_authorization_reply_unpack_server_msg():
    reply = TACACSAuthorizationReply.unpacked(
        six.b('\x01\x00\x08\x00\x00\x00amessage')
    )
    assert reply.server_msg == six.b('amessage')
    assert 'amessage' in str(reply)


def test_authorization_reply_unpack_data():
    reply = TACACSAuthorizationReply.unpacked(
        six.b('\x01\x00\x00\x00\x08\x00adata')
    )
    assert reply.data == six.b('adata')
    assert 'adata' in str(reply)


def test_authorization_reply_unpack_args():
    reply = TACACSAuthorizationReply.unpacked(
        six.b('\x01\x02\x00\x00\x00\x00\x04\x04arg1arg2')
    )
    assert reply.arg_cnt == 2
    assert reply.arguments == [b'arg1', b'arg2']


# test accounting object
def test_accounting_start_pack():
    username = 'user123'
    packed = TACACSAccountingStart(
        username, TAC_PLUS_ACCT_FLAG_START, TAC_PLUS_AUTHEN_METH_TACACSPLUS,
        TAC_PLUS_PRIV_LVL_MIN, TAC_PLUS_AUTHEN_TYPE_ASCII,
        [b"service=shell", b"cmd=show", b"cmdargs=version"],
    ).packed
    assert packed == six.b(
        '\x02'
        '\x06'              # tacacs_plus.flags.TAC_PLUS_AUTHEN_METH_TACACSPLUS \
        '\x00'              # tacacs_plus.flags.TAC_PLUS_PRIV_LVL_MIN \
        '\x01'              # tacacs_plus.flags.TAC_PLUS_AUTHEN_TYPE_ASCII \
        '\x01'              # tacacs_plus.flags.TAC_PLUS_AUTHEN_SVC_LOGIN \
        '\x07'              # username_len == 7 \
        '\x0b'              # port_len == 11 \
        '\x0d'              # rem_addr_len == 13 \
        '\x03'              # arg_len == 3
        '\x0d'              # arg_1_len == 13 (service=shell) \
        '\x08'              # arg_2_len == 8 (cmd=show) \
        '\x0f'              # arg_3_len == 14 (cmdargs=version) \
        'user123'           # username \
        'python_tty0'       # port \
        'python_device'     # rem_addr \
        'service=shell'     # arg_1 (service=shell) \
        'cmd=show'          # arg_2 (cmd=show) \
        'cmdargs=version'   # arg_3 (cmdargs=version)
    )


@pytest.mark.parametrize('raw, state, human_status', [
    (six.b('\x00\x00\x00\x00\x01\x00\x00'), 'valid', 'SUCCESS'),
    (six.b('\x00\x00\x00\x00\x02\x00\x00'), 'error', 'ERROR'),
    (six.b('\x00\x00\x00\x00\x21\x00\x00'), 'follow', 'FOLLOW'),
])
def test_accounting_reply_unpack(raw, state, human_status):
    reply = TACACSAccountingReply.unpacked(raw)
    assert getattr(reply, state) is True
    assert reply.human_status == human_status


def test_accounting_reply_unpack_server_msg():
    reply = TACACSAccountingReply.unpacked(
        six.b('\x08\x00\x00\x00\x01amessage')
    )
    assert reply.server_msg == six.b('amessage')
    assert 'amessage' in str(reply)


def test_accounting_reply_unpack_data():
    reply = TACACSAccountingReply.unpacked(
        six.b('\x00\x00\x08\x00\x01adata')
    )
    assert reply.data == six.b('adata')
    assert 'adata' in str(reply)


@pytest.mark.parametrize('family', [socket.AF_INET, socket.AF_INET6])
def test_v4_sock(family):
    try:
        client = TACACSClient('127.0.0.1', 49, None, family=family)
        client.sock
    except socket.error:
        pass
    assert client._sock.family == family


# test client send
@pytest.mark.parametrize(
    'packets, state',
    [
        [AUTHENTICATE_HEADER + six.b('\x06\x01\x00\x00\x00\x00\x00'), 'valid'],
        [AUTHENTICATE_HEADER + six.b('\x06\x02\x00\x00\x00\x00\x00'), 'invalid'],
        [AUTHENTICATE_HEADER + six.b('\x10\x05\x01\x00\n\x00\x00Password: '), 'getpass'],  # noqa
        [AUTHENTICATE_HEADER + six.b('\x06\x07\x00\x00\x00\x00\x00'), 'error'],
    ]
)
def test_client_socket_send(fake_socket, packets, state):
    body = TACACSAuthenticationStart(
        'user123',
        TAC_PLUS_AUTHEN_TYPE_ASCII
    )
    client = TACACSClient('127.0.0.1', 49, None, session_id=12345)
    client._sock = fake_socket
    packet = client.send(body, TAC_PLUS_AUTHEN)
    assert isinstance(packet, TACACSPacket)
    reply = TACACSAuthenticationReply.unpacked(packet.body)
    assert getattr(reply, state) is True

    # the first 12 bytes of the packet represent the header
    fake_socket.buff.seek(0)
    sent_header, sent_body = (
        fake_socket.buff.read(12), fake_socket.buff.read()
    )

    body_length = TACACSHeader.unpacked(sent_header).length
    assert len(sent_body) == body_length
    assert body.packed == sent_body


@pytest.mark.parametrize(
    'packets',
    [AUTHENTICATE_HEADER_WRONG + six.b('\x06\x07\x00\x00\x00\x00\x00')]
)
def test_client_socket_send_wrong_headers(fake_socket, packets):
    body = TACACSAuthenticationStart(
        'user123',
        TAC_PLUS_AUTHEN_TYPE_ASCII
    )
    client = TACACSClient('127.0.0.1', 49, None, session_id=12345)
    client._sock = fake_socket
    with pytest.raises(socket.error):
        client.send(body, TAC_PLUS_AUTHEN)


# test client.authenticate
@pytest.mark.parametrize(
    'packets',
    [
        AUTHENTICATE_HEADER + six.b('\x10\x05\x01\x00\n\x00\x00Password: ') +  # noqa getpass
        AUTHENTICATE_HEADER + six.b('\x06\x01\x00\x00\x00\x00\x00')  # auth_valid
    ]
)
def test_authenticate_ascii(fake_socket, packets):
    """
    client -> AUTHSTART (username)
              STATUS_GETPASS           <- server
    client -> AUTHCONTINUE (password)
              STATUS_PASS              <- server
    """
    client = TACACSClient('127.0.0.1', 49, None, session_id=12345)
    client._sock = fake_socket
    reply = client.authenticate('username', 'pass')
    assert reply.valid

    fake_socket.buff.seek(0)
    first_header = TACACSHeader.unpacked(fake_socket.buff.read(12))
    assert (first_header.version_max, first_header.version_min) == (12, 0)
    first_body = fake_socket.buff.read(first_header.length)
    assert TACACSAuthenticationStart(
        'username',
        TAC_PLUS_AUTHEN_TYPE_ASCII
    ).packed == first_body

    second_header = TACACSHeader.unpacked(fake_socket.buff.read(12))
    assert (first_header.version_max, first_header.version_min) == (12, 0)
    assert second_header.seq_no > first_header.seq_no

    second_body = fake_socket.buff.read()
    assert TACACSAuthenticationContinue('pass').packed == second_body


@pytest.mark.parametrize(
    'packets',
    [AUTHENTICATE_HEADER_V12_1 + six.b('\x06\x01\x00\x00\x00\x00\x00')]  # auth_valid
)
def test_authenticate_pap(fake_socket, packets):
    """
    client -> AUTHSTART (user+pass)
              STATUS_PASS              <- server
    """
    client = TACACSClient('127.0.0.1', 49, None, session_id=12345)
    client._sock = fake_socket
    reply = client.authenticate('username', 'pass',
                                authen_type=TAC_PLUS_AUTHEN_TYPE_PAP)
    assert reply.valid

    fake_socket.buff.seek(0)
    first_header = TACACSHeader.unpacked(fake_socket.buff.read(12))
    assert (first_header.version_max, first_header.version_min) == (12, 1)
    first_body = fake_socket.buff.read(first_header.length)
    assert TACACSAuthenticationStart(
        'username',
        TAC_PLUS_AUTHEN_TYPE_PAP,
        data=six.b('pass')
    ).packed == first_body


@pytest.mark.parametrize(
    'packets',
    [AUTHENTICATE_HEADER_V12_1 + six.b('\x06\x01\x00\x00\x00\x00\x00')]  # auth_valid
)
def test_authenticate_chap(fake_socket, packets):
    """
    client -> AUTHSTART user+md5challenge(pass)
              STATUS_PASS                         <- server
    """
    client = TACACSClient('127.0.0.1', 49, None, session_id=12345)
    client._sock = fake_socket
    reply = client.authenticate('username', 'pass',
                                authen_type=TAC_PLUS_AUTHEN_TYPE_CHAP,
                                chap_ppp_id='A',
                                chap_challenge='challenge')
    assert reply.valid

    fake_socket.buff.seek(0)
    first_header = TACACSHeader.unpacked(fake_socket.buff.read(12))
    assert (first_header.version_max, first_header.version_min) == (12, 1)
    first_body = fake_socket.buff.read(first_header.length)
    assert TACACSAuthenticationStart(
        'username',
        TAC_PLUS_AUTHEN_TYPE_CHAP,
        data=(
            six.b('A') +
            six.b('challenge') +
            md5(six.b('Apasschallenge')).digest()
        )
    ).packed == first_body


def test_authenticate_chap_ppp_id_required():
    client = TACACSClient('127.0.0.1', 49, None, session_id=12345)
    with pytest.raises(ValueError):
        client.authenticate('username', 'pass',
                            authen_type=TAC_PLUS_AUTHEN_TYPE_CHAP,
                            chap_challenge='challenge')


def test_authenticate_chap_ppp_id_length():
    client = TACACSClient('127.0.0.1', 49, None, session_id=12345)
    with pytest.raises(ValueError) as e:
        client.authenticate('username', 'pass',
                            authen_type=TAC_PLUS_AUTHEN_TYPE_CHAP,
                            chap_ppp_id='AA',
                            chap_challenge='challenge')
    assert 'chap_ppp_id must be a 1-byte string' in str(e)


def test_authenticate_chap_challenge_required():
    client = TACACSClient('127.0.0.1', 49, None, session_id=12345)
    with pytest.raises(ValueError):
        client.authenticate('username', 'pass',
                            authen_type=TAC_PLUS_AUTHEN_TYPE_CHAP,
                            chap_ppp_id='X')


def test_authenticate_chap_challenge_length():
    client = TACACSClient('127.0.0.1', 49, None, session_id=12345)
    with pytest.raises(ValueError) as e:
        client.authenticate('username', 'pass',
                            authen_type=TAC_PLUS_AUTHEN_TYPE_CHAP,
                            chap_ppp_id='A',
                            chap_challenge='X' * 256)
    assert 'chap_challenge may not be more 255 bytes' in str(e)


# test client.authorize
@pytest.mark.parametrize(
    'packets',
    [
        AUTHORIZE_HEADER + six.b('\x06\x01\x00\x00\x00\x00\x00')
    ]
)
def test_authorize_ascii(fake_socket, packets):
    client = TACACSClient('127.0.0.1', 49, None, session_id=12345)
    client._sock = fake_socket
    reply = client.authorize('username', arguments=[b"service=shell", b"cmd=show", b"cmdargs=version"])
    assert reply.valid

    fake_socket.buff.seek(0)
    first_header = TACACSHeader.unpacked(fake_socket.buff.read(12))
    assert (first_header.version_max, first_header.version_min) == (12, 0)
    first_body = fake_socket.buff.read(first_header.length)
    assert TACACSAuthorizationStart(
        'username', TAC_PLUS_AUTHEN_METH_TACACSPLUS, TAC_PLUS_PRIV_LVL_MIN,
        TAC_PLUS_AUTHEN_TYPE_ASCII, [b"service=shell", b"cmd=show", b"cmdargs=version"],
    ).packed == first_body


@pytest.mark.parametrize(
    'packets',
    [
        AUTHORIZE_HEADER + six.b('\x06\x01\x00\x00\x00\x00\x00')
    ]
)
def test_authorize_pap(fake_socket, packets):
    client = TACACSClient('127.0.0.1', 49, None, session_id=12345)
    client._sock = fake_socket
    reply = client.authorize('username', arguments=[b"service=shell", b"cmd=show", b"cmdargs=version"],
                             authen_type=TAC_PLUS_AUTHEN_TYPE_PAP)
    assert reply.valid

    fake_socket.buff.seek(0)
    first_header = TACACSHeader.unpacked(fake_socket.buff.read(12))
    assert (first_header.version_max, first_header.version_min) == (12, 0)
    first_body = fake_socket.buff.read(first_header.length)
    assert TACACSAuthorizationStart(
        'username', TAC_PLUS_AUTHEN_METH_TACACSPLUS, TAC_PLUS_PRIV_LVL_MIN,
        TAC_PLUS_AUTHEN_TYPE_PAP, [b"service=shell", b"cmd=show", b"cmdargs=version"],
    ).packed == first_body


@pytest.mark.parametrize(
    'packets',
    [
        AUTHORIZE_HEADER + six.b('\x06\x01\x00\x00\x00\x00\x00')
    ]
)
def test_authorize_chap(fake_socket, packets):
    client = TACACSClient('127.0.0.1', 49, None, session_id=12345)
    client._sock = fake_socket
    reply = client.authorize('username', arguments=[b"service=shell", b"cmd=show", b"cmdargs=version"],
                             authen_type=TAC_PLUS_AUTHEN_TYPE_CHAP)
    assert reply.valid

    fake_socket.buff.seek(0)
    first_header = TACACSHeader.unpacked(fake_socket.buff.read(12))
    assert (first_header.version_max, first_header.version_min) == (12, 0)
    first_body = fake_socket.buff.read(first_header.length)
    assert TACACSAuthorizationStart(
        'username', TAC_PLUS_AUTHEN_METH_TACACSPLUS, TAC_PLUS_PRIV_LVL_MIN,
        TAC_PLUS_AUTHEN_TYPE_CHAP, [b"service=shell", b"cmd=show", b"cmdargs=version"],
    ).packed == first_body


# test client.account
@pytest.mark.parametrize(
    'packets',
    [
        ACCOUNT_HEADER + six.b('\x06\x00\x00\x00\x00\x01\x00\x00')
    ]
)
def test_account_start(fake_socket, packets):
    client = TACACSClient('127.0.0.1', 49, None, session_id=12345)
    client._sock = fake_socket
    reply = client.account('username', TAC_PLUS_ACCT_FLAG_START,
                           arguments=[b"service=shell", b"cmd=show", b"cmdargs=version"])
    assert reply.valid

    fake_socket.buff.seek(0)
    first_header = TACACSHeader.unpacked(fake_socket.buff.read(12))
    assert (first_header.version_max, first_header.version_min) == (12, 0)
    first_body = fake_socket.buff.read(first_header.length)
    assert TACACSAccountingStart(
        'username', TAC_PLUS_ACCT_FLAG_START, TAC_PLUS_AUTHEN_METH_TACACSPLUS, TAC_PLUS_PRIV_LVL_MIN,
        TAC_PLUS_AUTHEN_TYPE_ASCII, [b"service=shell", b"cmd=show", b"cmdargs=version"],
    ).packed == first_body


@pytest.mark.parametrize(
    'packets',
    [
        AUTHORIZE_HEADER + six.b('\x12\x01\x01\x00\x00\x00\x00\x0bpriv-lvl=15')
    ]
)
def test_authorize_equal_priv_lvl(fake_socket, packets):
    client = TACACSClient('127.0.0.1', 49, None, session_id=12345)
    client._sock = fake_socket
    reply = client.authorize('username', arguments=[b"service=shell", b"cmd=show", b"cmdargs=version"],
                             authen_type=TAC_PLUS_AUTHEN_TYPE_PAP, priv_lvl=TAC_PLUS_PRIV_LVL_MAX)
    assert reply.valid, "the privilege level sent by the server is equal to the requested one (15)"


@pytest.mark.parametrize(
    'packets',
    [
        AUTHORIZE_HEADER + six.b('\x11\x01\x01\x00\x00\x00\x00\x0bpriv-lvl=1')
    ]
)
def test_authorize_lesser_priv_lvl(fake_socket, packets):
    client = TACACSClient('127.0.0.1', 49, None, session_id=12345)
    client._sock = fake_socket
    reply = client.authorize('username', arguments=[b"service=shell", b"cmd=show", b"cmdargs=version"],
                             authen_type=TAC_PLUS_AUTHEN_TYPE_PAP, priv_lvl=TAC_PLUS_PRIV_LVL_MAX)
    assert not reply.valid, "the privilege level sent by the server is less than the requested one (1 < 15)"



import random
import socket
import select
import logging
import contextlib
from hashlib import md5

import six

from .flags import (
    TAC_PLUS_MAJOR_VER, TAC_PLUS_MINOR_VER, TAC_PLUS_PRIV_LVL_MIN,
    TAC_PLUS_AUTHEN_TYPE_CHAP, TAC_PLUS_MINOR_VER_ONE, TAC_PLUS_AUTHEN_TYPE_PAP,
    TAC_PLUS_AUTHEN, TAC_PLUS_AUTHEN_TYPE_ASCII, TAC_PLUS_AUTHEN_METH_TACACSPLUS,
    TAC_PLUS_AUTHOR, TAC_PLUS_AUTHOR_STATUS_FAIL, TAC_PLUS_PRIV_LVL_MAX,
    TAC_PLUS_ACCT, TAC_PLUS_VIRTUAL_REM_ADDR, TAC_PLUS_VIRTUAL_PORT,
    TAC_PLUS_AUTHEN_STATUS_FAIL, TAC_PLUS_CONTINUE_FLAG_ABORT
)
from .packet import TACACSPacket, TACACSHeader
from .authentication import TACACSAuthenticationStart, TACACSAuthenticationContinue, TACACSAuthenticationReply
from .authorization import TACACSAuthorizationStart, TACACSAuthorizationReply
from .accounting import TACACSAccountingStart, TACACSAccountingReply

logger = logging.getLogger(__name__)


class TACACSClient(object):
    """
    A TACACS+ authentication client.
    https://datatracker.ietf.org/doc/draft-ietf-opsawg-tacacs

    An open source TACACS+ server daemon is available at
    http://www.shrubbery.net/tac_plus/
    """

    def __init__(self, host, port, secret, timeout=10, session_id=None,
                 family=socket.AF_INET, version_max=TAC_PLUS_MAJOR_VER,
                 version_min=TAC_PLUS_MINOR_VER):
        """
        :param host:        hostname of the TACACS+ server
        :param port:        port of the TACACS+ server, generally 49
        :param secret:      the secret key used to obfuscate packet bodies; can
                            be `None` to disable packet body obfuscation
        :param session_id:  a unique 32-bit int representing the session; if
                            left empty, one will be auto-generated
        :param version_max: TACACS+ major version number, 12
        :param version_min: TACACS+ minor version number, 0 or 1
        """
        self._sock = None
        self.host = host
        self.port = port
        self.secret = secret
        self.timeout = timeout
        self.version_max = version_max
        self.version_min = version_min
        self.family = family

        # session_id is an unsigned 32-bit int; unless it's provided, randomize
        self.session_id = session_id or random.randint(1, 2 ** 32 - 1)

    @property
    def version(self):
        return (self.version_max * 0x10) + self.version_min

    @property
    def sock(self):
        if not self._sock:  # pragma: nocover
            conn = (self.host, self.port)
            if self.family == socket.AF_INET6:
                # For AF_INET6 address family, a four-tuple (host, port,
                # flowinfo, scopeid) is used
                conn = (self.host, self.port, 0, 0)
            self._sock = socket.socket(self.family, socket.SOCK_STREAM)
            self._sock.setblocking(1)
            self._sock.settimeout(self.timeout)
            self._sock.connect(conn)
        return self._sock

    @contextlib.contextmanager
    def closing(self):
        try:
            self.sock
            yield
        finally:
            if self._sock:
                self._sock.close()
                self._sock = None

    def send(self, body, req_type, seq_no=1):
        """
        Send a TACACS+ message body

        :param body:     packed bytes, i.e., `struct.pack(...)`
        :param req_type: TAC_PLUS_AUTHEN,
                         TAC_PLUS_AUTHOR,
                         TAC_PLUS_ACCT
        :param seq_no:   The sequence number of the current packet.  The
                         first packet in a session MUST have the sequence
                         number 1 and each subsequent packet will increment
                         the sequence number by one.  Thus clients only send
                         packets containing odd sequence numbers, and TACACS+
                         servers only send packets containing even sequence
                         numbers.
        :return:         TACACSPacket
        :raises:         socket.timeout, socket.error
        """
        # construct a packet
        header = TACACSHeader(
            self.version,
            req_type,
            self.session_id,
            len(body.packed),
            seq_no=seq_no
        )
        packet = TACACSPacket(header, body.packed, self.secret)

        logger.debug('\n'.join([
            body.__class__.__name__,
            'sent header <%s>' % header,
            'sent body <%s>' % body,
        ]))
        self.sock.send(bytes(packet))

        readable, _, _ = select.select([self.sock], [], [], self.timeout)
        if readable:
            # TACACS+ header packets are always 12 bytes
            header_bytes = self.sock.recv(12)
            resp_header = TACACSHeader.unpacked(header_bytes)
            # If the reply header doesn't match, it's likely a non-TACACS+ TCP
            # service answering with data we don't antipicate.  Bail out.
            if any([
                resp_header.version_max != header.version_max,
                resp_header.type != header.type,
                resp_header.session_id != header.session_id
            ]):
                logger.error('\n'.join([
                    resp_header.__class__.__name__,
                    'recv header <%s>' % resp_header,
                    str(resp_header.packed)
                ]))
                raise socket.error

            # read the number of bytes specified in the response header
            body_bytes = six.b('')
            remaining = resp_header.length
            while remaining > 0:
                body_bytes += self.sock.recv(remaining)
                remaining = resp_header.length - len(body_bytes)
            return TACACSPacket(
                resp_header,
                body_bytes,
                self.secret
            )
        raise socket.timeout

    def authenticate(self, username, password, priv_lvl=TAC_PLUS_PRIV_LVL_MIN,
                     authen_type=TAC_PLUS_AUTHEN_TYPE_ASCII,
                     chap_ppp_id=None, chap_challenge=None,
                     rem_addr=TAC_PLUS_VIRTUAL_REM_ADDR, port=TAC_PLUS_VIRTUAL_PORT):
        """
        Authenticate to a TACACS+ server with a username and password.

        :param username:
        :param password:
        :param priv_lvl:
        :param authen_type:    TAC_PLUS_AUTHEN_TYPE_ASCII,
                               TAC_PLUS_AUTHEN_TYPE_PAP,
                               TAC_PLUS_AUTHEN_TYPE_CHAP
        :param chap_ppp_id:    PPP ID when authen_type == 'chap'
        :param chap_challenge: challenge value when authen_type == 'chap'
        :param rem_addr:       AAA request source, default to TAC_PLUS_VIRTUAL_REM_ADDR
        :param port:           AAA port, default to TAC_PLUS_VIRTUAL_PORT
        :return:               TACACSAuthenticationReply
        :raises:               socket.timeout, socket.error
        """
        start_data = six.b('')
        if authen_type in (TAC_PLUS_AUTHEN_TYPE_PAP,
                           TAC_PLUS_AUTHEN_TYPE_CHAP):
            self.version_min = TAC_PLUS_MINOR_VER_ONE

            if authen_type == TAC_PLUS_AUTHEN_TYPE_PAP:
                start_data = six.b(password)

            if authen_type == TAC_PLUS_AUTHEN_TYPE_CHAP:
                if not isinstance(chap_ppp_id, six.string_types):
                    raise ValueError('chap_ppp_id must be a string')
                if len(chap_ppp_id) != 1:
                    raise ValueError('chap_ppp_id must be a 1-byte string')
                if not isinstance(chap_challenge, six.string_types):
                    raise ValueError('chap_challenge must be a string')
                if len(chap_challenge) > 255:
                    raise ValueError('chap_challenge may not be more 255 bytes')
                start_data = (
                    six.b(chap_ppp_id) +
                    six.b(chap_challenge) +
                    md5(six.b(
                        chap_ppp_id + password + chap_challenge
                    )).digest()
                )
        with self.closing():
            packet = self.send(
                TACACSAuthenticationStart(username, authen_type, priv_lvl,
                                          start_data, rem_addr=rem_addr, port=port),
                TAC_PLUS_AUTHEN
            )
            reply = TACACSAuthenticationReply.unpacked(packet.body)
            logger.debug('\n'.join([
                reply.__class__.__name__,
                'recv header <%s>' % packet.header,
                'recv body <%s>' % reply
            ]))
            if authen_type == TAC_PLUS_AUTHEN_TYPE_ASCII and reply.getpass:
                packet = self.send(TACACSAuthenticationContinue(password),
                                   TAC_PLUS_AUTHEN,
                                   packet.seq_no + 1)
                reply = TACACSAuthenticationReply.unpacked(packet.body)
                logger.debug('\n'.join([
                    reply.__class__.__name__,
                    'recv header <%s>' % packet.header,
                    'recv body <%s>' % reply
                ]))
                if reply.flags == TAC_PLUS_CONTINUE_FLAG_ABORT:
                    reply.status = TAC_PLUS_AUTHEN_STATUS_FAIL

        return reply

    def authorize(self, username, arguments=[],
                  authen_type=TAC_PLUS_AUTHEN_TYPE_ASCII, priv_lvl=TAC_PLUS_PRIV_LVL_MIN,
                  rem_addr=TAC_PLUS_VIRTUAL_REM_ADDR, port=TAC_PLUS_VIRTUAL_PORT):
        """
        Authorize with a TACACS+ server.

        :param username:
        :param arguments:      The authorization arguments
        :param authen_type:    TAC_PLUS_AUTHEN_TYPE_ASCII,
                               TAC_PLUS_AUTHEN_TYPE_PAP,
                               TAC_PLUS_AUTHEN_TYPE_CHAP
        :param priv_lvl:       Minimal Required priv_lvl.
        :param rem_addr:       AAA request source, default to TAC_PLUS_VIRTUAL_REM_ADDR
        :param port:           AAA port, default to TAC_PLUS_VIRTUAL_PORT
        :return:               TACACSAuthenticationReply
        :raises:               socket.timeout, socket.error
        """
        with self.closing():
            packet = self.send(
                TACACSAuthorizationStart(username,
                                         TAC_PLUS_AUTHEN_METH_TACACSPLUS,
                                         priv_lvl, authen_type, arguments,
                                         rem_addr=rem_addr, port=port),
                TAC_PLUS_AUTHOR
            )
            reply = TACACSAuthorizationReply.unpacked(packet.body)
            logger.debug('\n'.join([
                reply.__class__.__name__,
                'recv header <%s>' % packet.header,
                'recv body <%s>' % reply
            ]))
            reply_arguments = dict([
                arg.split(six.b('='), 1)
                for arg in reply.arguments or []
                if arg.find(six.b('=')) > -1]
            )
            user_priv_lvl = int(reply_arguments.get(
                six.b('priv-lvl'), TAC_PLUS_PRIV_LVL_MAX))
            if user_priv_lvl < priv_lvl:
                reply.status = TAC_PLUS_AUTHOR_STATUS_FAIL
        return reply

    def account(self, username, flags, arguments=[],
                authen_type=TAC_PLUS_AUTHEN_TYPE_ASCII, priv_lvl=TAC_PLUS_PRIV_LVL_MIN,
                rem_addr=TAC_PLUS_VIRTUAL_REM_ADDR, port=TAC_PLUS_VIRTUAL_PORT):
        """
        Account with a TACACS+ server.

        :param username:
        :param flags:          TAC_PLUS_ACCT_FLAG_START,
                               TAC_PLUS_ACCT_FLAG_WATCHDOG,
                               TAC_PLUS_ACCT_FLAG_STOP
        :param arguments:      The authorization arguments
        :param authen_type:    TAC_PLUS_AUTHEN_TYPE_ASCII,
                               TAC_PLUS_AUTHEN_TYPE_PAP,
                               TAC_PLUS_AUTHEN_TYPE_CHAP
        :param priv_lvl:       Minimal Required priv_lvl.
        :param rem_addr:       AAA request source, default to TAC_PLUS_VIRTUAL_REM_ADDR
        :param port:           AAA port, default to TAC_PLUS_VIRTUAL_PORT
        :return:               TACACSAccountingReply
        :raises:               socket.timeout, socket.error
        """
        with self.closing():
            packet = self.send(
                TACACSAccountingStart(username, flags,
                                      TAC_PLUS_AUTHEN_METH_TACACSPLUS,
                                      priv_lvl, authen_type, arguments,
                                      rem_addr=rem_addr, port=port),
                TAC_PLUS_ACCT
            )
            reply = TACACSAccountingReply.unpacked(packet.body)
            logger.debug('\n'.join([
                reply.__class__.__name__,
                'recv header <%s>' % packet.header,
                'recv body <%s>' % reply
            ]))
        return reply



import struct

import six

from .flags import (
    TAC_PLUS_AUTHEN_SVC_LOGIN, TAC_PLUS_AUTHOR_STATUS_PASS_ADD, TAC_PLUS_AUTHOR_STATUS_PASS_REPL,
    TAC_PLUS_AUTHOR_STATUS_FAIL, TAC_PLUS_AUTHOR_STATUS_ERROR, TAC_PLUS_AUTHOR_STATUS_FOLLOW,
    TAC_PLUS_VIRTUAL_PORT, TAC_PLUS_VIRTUAL_REM_ADDR
)


class TACACSAuthorizationStart(object):

    def __init__(self, username, authen_method, priv_lvl, authen_type,
                 arguments, rem_addr=TAC_PLUS_VIRTUAL_REM_ADDR,
                 port=TAC_PLUS_VIRTUAL_PORT):
        self.username = username
        self.authen_method = authen_method
        self.priv_lvl = priv_lvl
        self.authen_type = authen_type
        self.service = TAC_PLUS_AUTHEN_SVC_LOGIN
        self.arguments = arguments
        self.rem_addr = rem_addr
        self.port = port

    @property
    def packed(self):
        #  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8
        # +----------------+----------------+----------------+----------------+
        # |  authen_method |    priv_lvl    |  authen_type   | authen_service |
        # +----------------+----------------+----------------+----------------+
        # |    user len    |    port len    |  rem_addr len  |    arg_cnt     |
        # +----------------+----------------+----------------+----------------+
        # |   arg 1 len    |   arg 2 len    |      ...       |   arg N len    |
        # +----------------+----------------+----------------+----------------+
        # |   user ...
        # +----------------+----------------+----------------+----------------+
        # |   port ...
        # +----------------+----------------+----------------+----------------+
        # |   rem_addr ...
        # +----------------+----------------+----------------+----------------+
        # |   arg 1 ...
        # +----------------+----------------+----------------+----------------+
        # |   arg 2 ...
        # +----------------+----------------+----------------+----------------+
        # |   ...
        # +----------------+----------------+----------------+----------------+
        # |   arg N ...
        # +----------------+----------------+----------------+----------------+

        # B = unsigned char
        # s = char[]
        username = six.b(self.username)
        rem_addr = six.b(self.rem_addr)
        port = six.b(self.port)
        arguments = self.arguments
        body = struct.pack(
            'B' * 8,
            self.authen_method,
            self.priv_lvl,
            self.authen_type,
            self.service,
            len(username),
            len(port),
            len(rem_addr),
            len(arguments),
        )
        for value in arguments:
            body += struct.pack('B', len(value))
        for value in (username, port, rem_addr):
            body += struct.pack('%ds' % len(value), value)
        for value in arguments:
            body += struct.pack('%ds' % len(value), value)
        return body

    def __str__(self):
        return ', '.join([
            'args: %s' % b','.join(self.arguments),
            'args_cnt: %d' % len(self.arguments),
            'authen_method: %s' % self.authen_method,
            'authen_type: %s' % self.authen_type,
            'authen_service: %s' % self.service,
            'port_len: %d' % len(self.port),
            'port: %s' % self.port,
            'priv_lvl: %s' % self.priv_lvl,
            'rem_addr_len: %d' % len(self.rem_addr),
            'user: %s' % self.username,
            'user_len: %d' % len(self.username),
        ])


class TACACSAuthorizationReply(object):

    def __init__(self, status, arg_cnt, server_msg, data, arguments):
        self.status = status
        self.arg_cnt = arg_cnt
        self.server_msg = server_msg
        self.data = data
        self.arguments = arguments
        self.flags = None

    @classmethod
    def unpacked(cls, raw):
        #  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8
        # +----------------+----------------+----------------+----------------+
        # |    status      |     arg_cnt    |         server_msg len          |
        # +----------------+----------------+----------------+----------------+
        # +            data len             |    arg 1 len   |    arg 2 len   |
        # +----------------+----------------+----------------+----------------+
        # |      ...       |   arg N len    |         server_msg ...
        # +----------------+----------------+----------------+----------------+
        # |   data ...
        # +----------------+----------------+----------------+----------------+
        # |   arg 1 ...
        # +----------------+----------------+----------------+----------------+
        # |   arg 2 ...
        # +----------------+----------------+----------------+----------------+
        # |   ...
        # +----------------+----------------+----------------+----------------+
        # |   arg N ...
        # +----------------+----------------+----------------+----------------+

        # B = unsigned char
        # !H = network-order (big-endian) unsigned short
        raw = six.BytesIO(raw)
        status, arg_cnt = struct.unpack('BB', raw.read(2))
        server_msg_len, data_len = struct.unpack('!HH', raw.read(4))
        args_lens = struct.unpack(
            'B' * arg_cnt, raw.read(arg_cnt)
        ) if arg_cnt else []
        server_msg = raw.read(server_msg_len)
        data = raw.read(data_len) if data_len else b''
        arguments = []
        for arg_len in args_lens:
            arg = raw.read(arg_len) if arg_len else b''
            arguments.append(arg)
        return cls(status, arg_cnt, server_msg, data, arguments)

    @property
    def valid(self):
        # authorization response is valid for both states:
        # * PASS_ADD -> cisco like, per command authorization
        # * PASS_REPL -> junos like, av-pairs authorization
        return self.status in (TAC_PLUS_AUTHOR_STATUS_PASS_ADD, TAC_PLUS_AUTHOR_STATUS_PASS_REPL)

    @property
    def invalid(self):
        return self.status == TAC_PLUS_AUTHOR_STATUS_FAIL

    @property
    def error(self):
        return self.status == TAC_PLUS_AUTHOR_STATUS_ERROR

    @property
    def reply(self):
        return self.status == TAC_PLUS_AUTHOR_STATUS_PASS_REPL

    @property
    def follow(self):
        return self.status == TAC_PLUS_AUTHOR_STATUS_FOLLOW

    @property
    def human_status(self):
        return {
            TAC_PLUS_AUTHOR_STATUS_PASS_ADD: 'PASS',
            TAC_PLUS_AUTHOR_STATUS_FAIL: 'FAIL',
            TAC_PLUS_AUTHOR_STATUS_PASS_REPL: 'REPL',
            TAC_PLUS_AUTHOR_STATUS_ERROR: 'ERROR',
            TAC_PLUS_AUTHOR_STATUS_FOLLOW: 'FOLLOW',
        }.get(self.status, 'UNKNOWN: %s' % self.status)

    def __str__(self):
        args = ', '.join([x.decode('utf-8') for x in self.arguments])
        return ', '.join([
            'args: %s' % args,
            'args_cnt: %d' % len(self.arguments),
            'data: %s' % self.data,
            'data_len: %d' % len(self.data),
            'server_msg: %s' % self.server_msg,
            'server_msg_len: %d' % len(self.server_msg),
            'status: %s' % self.human_status,
        ])



# versioning
TAC_PLUS_MAJOR_VER = 0xc
TAC_PLUS_MINOR_VER = 0x0
TAC_PLUS_MINOR_VER_ONE = 0x1

# types + actions
TAC_PLUS_AUTHEN = 0x01
TAC_PLUS_AUTHOR = 0x02
TAC_PLUS_AUTHEN_LOGIN = 0x01
TAC_PLUS_ACCT = 0x03

# services
TAC_PLUS_AUTHEN_SVC_NONE = 0x00
TAC_PLUS_AUTHEN_SVC_LOGIN = 0x01

# authentication types
TAC_PLUS_AUTHEN_TYPE_NOT_SET = 0x00
TAC_PLUS_AUTHEN_TYPE_ASCII = 0x01
TAC_PLUS_AUTHEN_TYPE_PAP = 0x02
TAC_PLUS_AUTHEN_TYPE_CHAP = 0x03
TAC_PLUS_AUTHEN_TYPES = {
    'ascii': TAC_PLUS_AUTHEN_TYPE_ASCII,
    'pap': TAC_PLUS_AUTHEN_TYPE_PAP,
    'chap': TAC_PLUS_AUTHEN_TYPE_CHAP,
}

# authentication flags
TAC_PLUS_CONTINUE_FLAG_ABORT = 0x01

# authentication statuses
TAC_PLUS_AUTHEN_STATUS_PASS = 0x01
TAC_PLUS_AUTHEN_STATUS_FAIL = 0x02
TAC_PLUS_AUTHEN_STATUS_GETDATA = 0x03
TAC_PLUS_AUTHEN_STATUS_GETUSER = 0x04
TAC_PLUS_AUTHEN_STATUS_GETPASS = 0x05
TAC_PLUS_AUTHEN_STATUS_RESTART = 0x06
TAC_PLUS_AUTHEN_STATUS_ERROR = 0x07
TAC_PLUS_AUTHEN_STATUS_FOLLOW = 0x21

# priveleges
TAC_PLUS_PRIV_LVL_MIN = 0x00
TAC_PLUS_PRIV_LVL_MAX = 0x0F

# authorization statuses
TAC_PLUS_AUTHOR_STATUS_PASS_ADD = 0x01
TAC_PLUS_AUTHOR_STATUS_PASS_REPL = 0x02
TAC_PLUS_AUTHOR_STATUS_FAIL = 0x10
TAC_PLUS_AUTHOR_STATUS_ERROR = 0x11
TAC_PLUS_AUTHOR_STATUS_FOLLOW = 0x21

# authentication methods
TAC_PLUS_AUTHEN_METH_NOT_SET = 0x00
TAC_PLUS_AUTHEN_METH_NONE = 0x01
TAC_PLUS_AUTHEN_METH_KRB5 = 0x02
TAC_PLUS_AUTHEN_METH_LINE = 0x03
TAC_PLUS_AUTHEN_METH_ENABLE = 0x04
TAC_PLUS_AUTHEN_METH_LOCAL = 0x05
TAC_PLUS_AUTHEN_METH_TACACSPLUS = 0x06
TAC_PLUS_AUTHEN_METH_GUEST = 0x08
TAC_PLUS_AUTHEN_METH_RADIUS = 0x10
TAC_PLUS_AUTHEN_METH_KRB4 = 0x11
TAC_PLUS_AUTHEN_METH_RCMD = 0x20

# accounting methods
TAC_PLUS_ACCT_FLAG_START = 0x02
TAC_PLUS_ACCT_FLAG_STOP = 0x04
TAC_PLUS_ACCT_FLAG_WATCHDOG = 0x08
TAC_PLUS_ACCT_FLAGS = {
    'start': TAC_PLUS_ACCT_FLAG_START,
    'stop': TAC_PLUS_ACCT_FLAG_STOP,
    'update': TAC_PLUS_ACCT_FLAG_WATCHDOG
}

# accounting statuses
TAC_PLUS_ACCT_STATUS_SUCCESS = 0x01
TAC_PLUS_ACCT_STATUS_ERROR = 0x02
TAC_PLUS_ACCT_STATUS_FOLLOW = 0x21

# port and rem_addr, useful inside tac_plus accounting logs
TAC_PLUS_VIRTUAL_PORT = "python_tty0"
TAC_PLUS_VIRTUAL_REM_ADDR = "python_device"



from setuptools import setup, find_packages

version = '2.6'

setup(
    name='tacacs_plus',
    version=version,
    description="A client for TACACS+ authentication",
    long_description=None,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    keywords='tacas tacacs+ tacacs_plus auth authentication pap chap',
    author='Ryan Petrello',
    author_email='ryan@ryanpetrello.com',
    url='http://github.com/ansible/tacacs_plus',
    license='BSD',
    packages=find_packages(exclude=['examples']),
    install_requires=['six'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'pytest-cov'],
    include_package_data=True,
    scripts=['bin/tacacs_client'],
    zip_safe=False
)



import struct

import six

from .flags import (
    TAC_PLUS_PRIV_LVL_MIN, TAC_PLUS_AUTHEN_LOGIN, TAC_PLUS_AUTHEN_SVC_LOGIN,
    TAC_PLUS_AUTHEN_STATUS_PASS, TAC_PLUS_AUTHEN_STATUS_FAIL, TAC_PLUS_AUTHEN_STATUS_ERROR,
    TAC_PLUS_AUTHEN_STATUS_GETPASS, TAC_PLUS_VIRTUAL_PORT, TAC_PLUS_VIRTUAL_REM_ADDR
)


class TACACSAuthenticationStart(object):

    def __init__(self, username, authen_type, priv_lvl=TAC_PLUS_PRIV_LVL_MIN,
                 data=six.b(''), rem_addr=TAC_PLUS_VIRTUAL_REM_ADDR,
                 port=TAC_PLUS_VIRTUAL_PORT):
        self.username = username
        self.action = TAC_PLUS_AUTHEN_LOGIN
        self.priv_lvl = priv_lvl
        self.authen_type = authen_type
        self.service = TAC_PLUS_AUTHEN_SVC_LOGIN
        self.data = data
        self.rem_addr = rem_addr
        self.port = port

    @property
    def packed(self):
        # 1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8
        #
        # +----------------+----------------+----------------+----------------+
        # |    action      |    priv_lvl    |  authen_type   |     service    |
        # +----------------+----------------+----------------+----------------+
        # |    user len    |    port len    |  rem_addr len  |    data len    |
        # +----------------+----------------+----------------+----------------+
        # |    user ...
        # +----------------+----------------+----------------+----------------+
        # |    port ...
        # +----------------+----------------+----------------+----------------+
        # |    rem_addr ...
        # +----------------+----------------+----------------+----------------+
        # |    data...
        # +----------------+----------------+----------------+----------------+

        # B = unsigned char
        # s = char[]
        username = six.b(self.username)
        rem_addr = six.b(self.rem_addr)
        port = six.b(self.port)
        data = self.data
        body = struct.pack(
            'B' * 8,
            self.action,
            self.priv_lvl,
            self.authen_type,
            self.service,
            len(username),
            len(port),
            len(rem_addr),
            len(data),
        )
        for value in (username, port, rem_addr, data):
            body += struct.pack('%ds' % len(value), value)
        return body

    def __str__(self):
        return ', '.join([
            'action: %s' % self.action,
            'authen_type: %s' % self.authen_type,
            'authen_service: %s' % self.service,
            'data: %s' % self.data,
            'data_len: %d' % len(self.data),
            'priv_lvl: %s' % self.priv_lvl,
            'port: %s' % self.port,
            'port_len: %d' % len(self.port),
            'rem_addr: %s' % self.rem_addr,
            'rem_addr_len: %d' % len(self.rem_addr),
            'user: %s' % self.username,
            'user_len: %d' % len(self.username)
        ])


class TACACSAuthenticationContinue(object):
    def __init__(self, password, data=six.b(''), flags=0):
        self.password = password
        self.data = data
        self.flags = flags

    @property
    def packed(self):
        # 1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8
        #
        # +----------------+----------------+----------------+----------------+
        # |          user_msg len           |            data len             |
        # +----------------+----------------+----------------+----------------+
        # |     flags      |  user_msg ...
        # +----------------+----------------+----------------+----------------+
        # |    data ...
        # +----------------+

        # B = unsigned char
        # !H = network-order (big-endian) unsigned short
        # s = char[]
        password = six.b(self.password)
        data = self.data
        return (
            struct.pack('!H', len(password)) +
            struct.pack('!H', len(data)) +
            struct.pack('B', self.flags) +
            struct.pack('%ds' % len(password), password) +
            struct.pack('%ds' % len(data), data)
        )

    def __str__(self):
        return ', '.join([
            'data_len: 0',
            'flags: 0',
            'user_msg: %s' % ('*' * len(self.password)),
            'user_msg_len: %s' % len(self.password)
        ])


class TACACSAuthenticationReply(object):

    def __init__(self, status, flags, server_msg, data):
        self.status = status
        self.flags = flags
        self.server_msg = server_msg
        self.data = data
        self.arguments = []

    @classmethod
    def unpacked(cls, raw):
        # 1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8  1 2 3 4 5 6 7 8
        #
        # +----------------+----------------+----------------+----------------+
        # |     status     |      flags     |        server_msg len           |
        # +----------------+----------------+----------------+----------------+
        # |           data len              |        server_msg ...
        # +----------------+----------------+----------------+----------------+
        # |           data ...
        # +----------------+----------------+

        # B = unsigned char
        # !H = network-order (big-endian) unsigned short
        raw = six.BytesIO(raw)
        status, flags = struct.unpack('BB', raw.read(2))
        server_msg_len, data_len = struct.unpack('!HH', raw.read(4))
        server_msg = raw.read(server_msg_len)
        data = raw.read(data_len)
        return cls(status, flags, server_msg, data)

    @property
    def valid(self):
        return self.status == TAC_PLUS_AUTHEN_STATUS_PASS

    @property
    def invalid(self):
        return self.status == TAC_PLUS_AUTHEN_STATUS_FAIL

    @property
    def error(self):
        return self.status == TAC_PLUS_AUTHEN_STATUS_ERROR

    @property
    def getpass(self):
        return self.status == TAC_PLUS_AUTHEN_STATUS_GETPASS

    @property
    def human_status(self):
        return {
            TAC_PLUS_AUTHEN_STATUS_PASS: 'PASS',
            TAC_PLUS_AUTHEN_STATUS_FAIL: 'FAIL',
            TAC_PLUS_AUTHEN_STATUS_GETPASS: 'GETPASS',
            TAC_PLUS_AUTHEN_STATUS_ERROR: 'ERROR'
        }.get(self.status, 'UNKNOWN: %s' % self.status)

    def __str__(self):
        return ', '.join([
            'data: %s' % self.data,
            'data_len: %d' % len(self.data),
            'flags: %s' % self.flags,
            'server_msg: %s' % self.server_msg,
            'server_msg_len: %d' % len(self.server_msg),
            'status: %s' % self.human_status
        ])
